/*
 * Harmonic Oscillator: i*u_t = -(1/2)*u_xx + (x^2/2)*u
 */

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <complex>

typedef std::complex<double> Complex;
const Complex I(1,0), J(0,1);

#define PI  3.14159265358979323846

/* global staff */
#define IMAX 500
#define NBUF 2
Complex u[NBUF][IMAX+1];

/* window size adjusted to IMAX */
int width = IMAX, height = IMAX/2;

/* solving scheme */
void CentDif(void);
void Ceyley(void);
void (*pde)() = CentDif;

/* solving parameters */
#define L (30.0) /* length of solving region */
double  A = 1;   /* amplitude */
double  a = 0;   /* center of packet */
double  k = 7;   /* average momentum */
double  r = 1;   /* dt mag */
int     N = 1;   /* musk for quantum num of envelope */

/* flags */
int initialize = 1;
int Dirichlet  = 1;
int pause      = 0;

/* drawing parameters */
int draw_step = 10;
double   ymag = 0.3;

/***************************************************
 * solve PDE: i*u_t = -(1/2)*u_xx + (x^2/2)*u
 --------------------------------------------------*/
#define dx   ( L/IMAX )
#define x(i) ( dx*(i) - .5*L )
#define dt   ( dx*dx * .4 )    /* critical at .42-.43*/
#define i(D) ( (i+(D)+IMAX) % IMAX )
#define n(D) ( (n+(D)+NBUF) % NBUF )

int bitcount(unsigned n)
{ int b; for (b = 0; n != 0; n >>= 1) if (n & 01) b++; return b; }

/*
 * Center difference scheme on time
 */
void CentDif(void)
{
    static int n = 0;

    /* initial value */
    if (initialize) {
	initialize = 0;
	N = (N == 0) ? 1 : N;
	for (int i = 0; i <= IMAX; i++) {
	    double Gauss = exp(-(x(i)-a)*(x(i)-a)/(2*A*A))/sqrt(sqrt(PI))/A
		* (((N & 1)!=0) * 1.0 +
		   ((N & 2)!=0) * x(i) +
		   ((N & 4)!=0) * (x(i)*x(i) - .5) +
		   ((N & 8)!=0) * (x(i)*x(i)*x(i) - 12./8.*x(i))
		   ) / bitcount(N);
	    u[n(0)][i] = Gauss * exp(J*x(i)*k);
	}
	for (int i=0; i<=IMAX; i++) {
	    u[n(-1)][i]
	 =
	    u[n(0)][i]
	  - J*(dt/dx/dx/2)*(u[n(0)][i(+1)] - 2.*u[n(0)][i] + u[n(0)][i(-1)])
          + J*(dt*x(i)*x(i)/2)*u[n(0)][i];
	}
    }

    /* evolve */
    for (int i=0; i<=IMAX; i++) {
	u[n(+1)][i]
     = 
	u[n(-1)][i]
      + J*(dt/dx/dx)*(u[n(0)][i(+1)] - 2.*u[n(0)][i] + u[n(0)][i(-1)])
      - J*(dt*x(i)*x(i))*u[n(0)][i];
    }

    /* boundary condition */
    if (Dirichlet) u[n(+1)][0] = u[n(+1)][IMAX] = 0;

    if (n++% draw_step == 0) glutPostRedisplay();
}

/*
 * Ceyley form: Crank Nicolson difference scheme
 */
void Ceyley(void)
{
    void tridiag(Complex[], Complex[], Complex[], int);
    static int n = 0; /* time count */

    /* initial value */
    if (initialize) {
	initialize = 0;
	N = (N == 0) ? 1 : N;
	for (int i = 0; i <= IMAX; i++) {
	    double Gauss = exp(-(x(i)-a)*(x(i)-a)/(2*A*A))/sqrt(sqrt(PI))/A
		* (((N & 1)!=0) * 1.0 +
		   ((N & 2)!=0) * x(i) +
		   ((N & 4)!=0) * (x(i)*x(i) - .5) +
		   ((N & 8)!=0) * (x(i)*x(i)*x(i) - 12./8.*x(i))
		   ) / bitcount(N);
	    u[n(0)][i] = Gauss * exp(J*x(i)*k);
	}
    }

    /* evolve */
    static double  V[IMAX+1];
    static Complex D[IMAX+1];
    static Complex R[IMAX+1];
    for (int i = 0; i <= IMAX; i++) {
	V[i] = .5*x(i)*x(i);
	D[i] =  J*(4*dx*dx/dt) - 2. - 2*dx*dx*V[i];
	R[i] = (J*(4*dx*dx/dt) + 2. + 2*dx*dx*V[i])*u[n(0)][i]
	      - u[n(0)][i(-1)] - u[n(0)][i(+1)];
    }
    tridiag(u[n(+1)], D, R, IMAX);

    /* boundary condition */
    if (Dirichlet) u[n(+1)][0] = u[n(+1)][IMAX] = 0;

    if ((n++ % draw_step) == 0) glutPostRedisplay();
}

void tridiag(Complex x[], Complex d[], Complex r[], int N)
{
    static Complex *t = 0;
    static int last_n = 0;

    if (N != last_n) {
	last_n = N;
	if (t)  free(t);
	t = (Complex *)malloc(sizeof(Complex)*N);
    }
    if (d[0] == 0.0) { fprintf(stderr, "d[0] == 0\n"); exit(1); }
    t[0] = 1./d[0];
    x[0] = r[0]*t[0];
    for (int i = 1; i < N; i++) {
	t[i] = 1./(d[i] - t[i-1]);
	x[i] = (r[i] - x[i-1])*t[i];
    }
    for (int i = N-2; i >= 0; i--)
	x[i] = x[i] - x[i+1]*t[i]; 
}

/*******************************
 * display solution
 ------------------------------*/
void draw(Complex *fn)
//
#define X(i)   ( width/IMAX*(i) )
#define Y(y)   ( ((y)*ymag + .5)*height )
{
    /* axis */
    glColor3f(.2,.2,.3);
    glBegin(GL_LINES);
    glVertex2d(0, Y(0)); glVertex2d(width, Y(0));
    glEnd();

    /* potential */
    glColor3f(.2,.2,.4);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i), Y(.2/2*(x(i)*x(i)-k*k-a*a-1-2*log(N)/log(2))));
    glEnd();

    /* envelope */
    glColor3f(.0,.4,.3);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i), Y(abs(fn[i])));
    glEnd();
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i), Y(-abs(fn[i])));
    glEnd();

    /* sol */
    glColor3f(.0,.9,.0);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i),Y(real(fn[i])));
    glEnd();
}

/*****************************************
 * glut call-back
 ----------------------------------------*/
void display(void) 
{
    glClearColor(.0,.0,.0,.0);
    glClear(GL_COLOR_BUFFER_BIT);
    draw(&u[0][0]); 
    glutSwapBuffers();
}

void idle(void) { if (!pause) pde(); }

void reshape(int w, int h)
{
    glutReshapeWindow(width, (height = (h/2)*2));
    glViewport(0,0,width, height);
    glMatrixMode(GL_PROJECTION); {
        glLoadIdentity();
        gluOrtho2D(0., width - 1., 0., height - 1.);
    }
    pause = 0;
    glutPostRedisplay();
}

void action(int key)
{
    static int cur_num = 7;

    if (key >= '0' && key <= '9') { cur_num = key - '0'; return; }
    if (key >= 'a' && key <= 'f') { cur_num = key - 'a' + 10; return; }
    switch (key) {
    case 'q': case '': exit(0);
    case 'p':  pause = !pause; break;
    case 'r':  initialize = 1; pause = 0; break;
    case 'n':  N = cur_num; initialize = 1; break;
    case 'k':  k = cur_num; initialize = 1; break;
    case ',':  pde = CentDif; break;
    case '.':  pde = Ceyley; break;
    case '/':  draw_step = draw_step/2 + 1; break;
    case '*':  draw_step = draw_step*2; break;
    case '+':  ymag *= 1.2; break;
    case ';':  ymag /= 1.2; break;
    default :  break;
    }
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y) { action((int) key); }

void qn_select(int value) { N = value; initialize = 1; }
void k_select(int value)  { k = value; initialize = 1; }
void scheme_select(int value) { pde = (value) ? Ceyley : CentDif; }

void make_menu(void)
{
    int qn, k, scheme;
    qn = glutCreateMenu(qn_select); {
        glutAddMenuEntry("0",  1);
        glutAddMenuEntry("1",  2);
        glutAddMenuEntry("2",  4);
        glutAddMenuEntry("3",  8);
    }
    k = glutCreateMenu(k_select); {
        glutAddMenuEntry(" 0", 0);
        glutAddMenuEntry(" 2", 2);
        glutAddMenuEntry(" 4", 4);
	glutAddMenuEntry(" 7", 7);
	glutAddMenuEntry("10",10);
    }
    scheme = glutCreateMenu(scheme_select); {
        glutAddMenuEntry("Center-difference", 0);
        glutAddMenuEntry("Crank-Nicolson",   1);
    }
    glutCreateMenu(action); {
        glutAddMenuEntry("Pause/Continue",       'p');
        glutAddMenuEntry("Restart",              'r');
	glutAddSubMenu("Quantum Num of envelope", qn);
        glutAddSubMenu("Momentum of guide wave",  k );
        glutAddSubMenu("Solving Scheme",      scheme);
        glutAddMenuEntry("Slow down",            '/');
        glutAddMenuEntry("Speed up",             '*');
        glutAddMenuEntry("Quit",                 'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*****************************************
 * main
 ----------------------------------------*/
int main(int argc, char** argv)
{
    /* init glut */
    glutInit (&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutCreateWindow(argv[0]);

    /* call-back */
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutKeyboardFunc(keyboard);
    make_menu();

    /* 2D-setup */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0., width - 1., 0., height - 1.);

    glutMainLoop ();
    return 0; 
}
