/*
 * KdV equation: u_t + 6*u*u_x + u_xxx = 0
 */

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <math.h>

/* global staff */
#define IMAX 500
#define NBUF 2
static  double u[NBUF][IMAX+1];

/* window size */
int width = IMAX, height = IMAX/2;

/* solving equation */
void KdV(void);
void (*pde)() = KdV;

/* global parameters */
#define L  (100.0)   /* period (length) */
double  A0 =  3.0;   /* amplitude */
double  N0 = 10.0;   /* = N[k] for k-soliton */
double  N[]= {1, 1, 3, 6, 10, 15, 21, 28, 100, 300, 500};

/* flags */
int initialize = 1;
int pause = 0;

/* drawing parameters */
int draw_step = 10;
int draw_mode =  0;
double ymag   = 1.;

/*****************************************
 * solve KdV:  u_t + 6*u*u_x + u_xxx = 0
 ----------------------------------------*/
void KdV(void)
//
#define dx   ( L/IMAX )
#define x(i) ( dx*(i) - .5*L )
#define dt   ( dx*dx*dx * .35 )  /* critical at .38-.39 */
#define i(D) ( (i+(D)+IMAX) % IMAX )
#define n(D) ( (n+(D)+NBUF) % NBUF )
{
    static int n = 0;

    /* initial value */
    if (initialize) {
	initialize = 0;
	for (int i = 0; i <= IMAX; i++) {
	    double A = A0/N0;
	    double Sech = 1/cosh(sqrt(A)/2*x(i));
	    u[n(0)][i] = A/2*Sech*Sech*N0;
	}
	for (int i = 0; i <= IMAX; i++) {
	    u[n(-1)][i]
	=
	    u[n(0)][i]
	+ .5*6*(dt/dx)*(u[n(0)][i(+1)] + u[n(0)][i] + u[n(0)][i(-1)])/3
	    * (u[n(0)][i(+1)] - u[n(0)][i(-1)])
	+ .5*(dt/dx/dx/dx)*(u[n(0)][i(+2)]
			    - 2*u[n(0)][i(+1)] + 2*u[n(0)][i(-1)]
			    - u[n(0)][i(-2)]);
	}
    }
    /* evolve */
    for (int i = 0; i <= IMAX; i++) {
	u[n(+1)][i]
     = 
	u[n(-1)][i]
      - 6*(dt/dx)*(u[n(0)][i(+1)] + u[n(0)][i] + u[n(0)][i(-1)])/3
	  * (u[n(0)][i(+1)] - u[n(0)][i(-1)])
      - (dt/dx/dx/dx)*(u[n(0)][i(+2)]
			 - 2*u[n(0)][i(+1)] + 2*u[n(0)][i(-1)]
			 - u[n(0)][i(-2)]);
    }

    if (n++ % draw_step == 0) glutPostRedisplay();
}

/*****************************************
 * display solution
 ----------------------------------------*/
void draw(double *fn)
//
#define X(i)   ( width/IMAX*(i) )
#define Y(y)   ( (.5*(y)/A0*ymag + .25)*height )
{
    glClearColor(.0,.0,.0,.0);
    glClear(GL_COLOR_BUFFER_BIT);

    /* axis */
    glColor3f(.2,.2,.3);
    glBegin(GL_LINES);
    glVertex2d(0, Y(0)); glVertex2d(width, Y(0));
    glEnd();

    /* sol */
    glColor3f(.0,.9,.0);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i),Y(fn[i]));
    glEnd();
}

void draw1(double *fn)
//
#define X1(i)   ( width/IMAX*(i) )  /* must be integer */
#define Y1(y)   ( 30*(y)/A0*ymag + 8 + (count/skip)*3 )
{
    static double lowerhorizon[1024], upperhorizon[1024];
    static int visible = 0, visible1 = 0;
    static int count = 0;
    const  int skip  = 5;

    if (count == 0) {
	glClear(GL_COLOR_BUFFER_BIT);
        for (int i = 0; i < 1024; i++) {
            upperhorizon[i] = -1e23;
	    lowerhorizon[i] = +1e23;
        }
	visible = visible1 = 0;
    }
    if ((++count%skip) != 0) { return ; }

    glBegin(GL_LINES);
    for (int i = 0; i <= IMAX; i++) {
        double y = Y1(fn[i]);
        visible = 0;
        if (y < lowerhorizon[X1(i)]) {
            lowerhorizon[X1(i)] = y;
            // visible = 1;
        }
	if (y > upperhorizon[X1(i)]) {
            upperhorizon[X1(i)] = y;
            visible = 1;
        }
        if (visible1 && visible) {
            glVertex2d(X1(i-1), Y1(fn[i-1]));
	    glVertex2d(X1(i), Y1(fn[i]));
	}
	visible1 = visible;
    }
    glEnd();

    if ((count % ((height - 30)/3*skip)) == 0) {
	count = 0;
	pause = 1;
    }
}

/*****************************************
 * glut call-back
 ----------------------------------------*/
void display(void) 
{
    if (draw_mode == 0)
	draw(&u[0][0]);
    else
 	draw1(&u[0][0]); 
    glutSwapBuffers();
}

void idle(void) { if (!pause) pde(); }

void reshape(int w, int h)
{
    glutReshapeWindow(width, (height = (h/2)*2));
    glViewport(0,0,width, height);
    glMatrixMode(GL_PROJECTION); {
	glLoadIdentity();
	gluOrtho2D(0., width - 1., 0., height - 1.);
    }
    pause = 0;
    glutPostRedisplay();
}

void action(int key)
{
    pause = 0;
    if (key >= '0' && key <= '9') {
	N0 = N[key-'0']; initialize = 1; return;
    }
    switch (key) {
    case 'q': exit(0);
    case 'r': initialize = 1; break;
    case 'p': pause = 1; break;
    case '/': draw_step = draw_step/2 + 1; break;
    case '*': draw_step = draw_step*2; break;
    case '+': ymag *= 1.2; break;
    case ';': ymag /= 1.2; break;
    case '-': ymag /= 1.2; break;
    case 'a': draw_mode = !draw_mode; break;
    default:  break;
    }
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)  { action((int) key); }
void mouse(int state, int button, int x, int y) { action('c'); }

void make_menu(void)
{
    int N0;
    N0 = glutCreateMenu(action); {
        glutAddMenuEntry(" 1", '1');
        glutAddMenuEntry(" 2", '2');
        glutAddMenuEntry(" 3", '3');
        glutAddMenuEntry(" 4", '4');
        glutAddMenuEntry(" 5", '5');
        glutAddMenuEntry(" 6", '6');
        glutAddMenuEntry(" 7", '7');
        glutAddMenuEntry("many", '9');
    }
    glutCreateMenu(action); {
        glutAddMenuEntry("Continue",     'c');
        glutAddMenuEntry("Pause",        'p');
        glutAddMenuEntry("Another View", 'a');
	glutAddSubMenu("Num Soliton: N=", N0);
        glutAddMenuEntry("Restart",      'r');
        glutAddMenuEntry("Slow down",    '/');
        glutAddMenuEntry("Speed up",     '*');
        glutAddMenuEntry("Quit",         'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*****************************************
 * main
 ----------------------------------------*/
int main(int argc, char** argv)
{
    /* init glut */
    glutInit (&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutCreateWindow(argv[0]);

    /* call-back */
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutKeyboardFunc(keyboard);
    glutMouseFunc(mouse);
    make_menu();

    /* 2D-setup */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0., width - 1., 0., height - 1.);

    glutMainLoop ();
    return 0; 
}
