/*
 * Non-linear Schrodinger eq: i*u_t = -(1/2)*u_xx - (|u|^2)*u
 */

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <complex>

typedef std::complex<double> Complex;
const   Complex I(1,0), J(0,1);

/* global staff */
#define IMAX 500
#define NBUF 2
static Complex u[NBUF][IMAX+1];

/* window size adjusted to xgrid */
int width = IMAX, height = IMAX/2;

/* solving equation */
void NLS(void);
void (*pde)(void) = NLS;

/* equation parameters */
#define L   (20.0)   /* period (length) */
double  A0 =  3.0;   /* amplitude0 */
double  A1 =  2.0;   /* amplitude1 */
double  k0 =  8.0;   /* momentum0 */
double  k1 = 12.5;   /* momentum1 */

/* flags */
int initialize = 1;
int pause      = 0;

/* drawing parameters */
int draw_step = 10;
double ymag   = .25;

/*****************************************
 * solve PDE
 ----------------------------------------*/
void NLS(void)
//
#define dx   ( L/IMAX )
#define dt   ( dx*dx*.45 )    /* critial at 0.50-0.51 */
#define x(i) ( dx*(i) - .5*L )
#define i(D) ( (i+(D)+IMAX) % IMAX )
#define n(D) ( (n+(D)+NBUF) % NBUF )
{
    static int n = 0;

    /* initial value */
    if (initialize) {
	initialize = 0;
	for (int i = 0; i <= IMAX; i++) {
	    double Sech0 = A0 / cosh( A0 * (x(i) - .1*L) );
	    double Sech1 = A1 / cosh( A1 * (x(i) + .2*L) );
	    u[n(0)][i] = Sech0*exp(J*x(i)*(-k0)) + Sech1*exp(J*x(i)*k1);
	}
	for (int i=0; i<=IMAX; i++) {
	    double au = abs(u[n(0)][i]);
	    u[n(-1)][i]
	 =
	    u[n(0)][i]
	  - J*(dt/dx/dx)/2.*(u[n(0)][i(+1)] - 2.*u[n(0)][i] + u[n(0)][i(-1)])
	  - J*dt*(au*au)*u[n(0)][i];
	}
    }

    /* evolve */
    for (int i = 0; i <= IMAX; i++) {
	double au = abs(u[n(0)][i]);
	u[n(+1)][i]
     = 
	u[n(-1)][i]
      + J*(dt/dx/dx)*(u[n(0)][i(+1)] - 2.*u[n(0)][i] + u[n(0)][i(-1)])
      + J*2.*dt* (au*au)*u[n(0)][i];
    }
    if (n++ % draw_step == 0) glutPostRedisplay();
}

/*******************************
 * display solution
 ------------------------------*/
void draw(Complex *fn)
//
#define X(i)   ( width/IMAX*(i) )
#define Y(y)   ( ((y)/A0*ymag + .5)*height )
{
    /* axis */
    glColor3f(.2,.2,.3);
    glBegin(GL_LINES);
    glVertex2d(0, Y(0)); glVertex2d(width, Y(0));
    glEnd();

    /* envelope */
    glColor3f(.0,.4,.3);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i), Y(abs(fn[i])));
    glEnd();
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i), Y(-abs(fn[i])));
    glEnd();

    /* solution */
    glColor3f(.0,.9,.0);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i), Y(real(fn[i])));
    glEnd();
}

/*****************************************
 * glut call-back
 ----------------------------------------*/
void display(void) 
{
    glClearColor(.0,.0,.0,.0);
    glClear(GL_COLOR_BUFFER_BIT);
    draw(&u[0][0]); 
    glutSwapBuffers();
}

void idle(void) { if (!pause) pde(); }

void reshape(int w, int h)
{
    glutReshapeWindow(width, (height = (h/2)*2));
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION); {
        glLoadIdentity();
	gluOrtho2D(0., width - 1., 0., height - 1.);
    }
    pause = 0;
    glutPostRedisplay();
}

void action(int key)
{
    static int cur_num = 7;

    if (key >= '0' && key <= '9') { cur_num = key - '0'; return; }
    switch (key) {
    case 'q': case '': exit(0);
    case 'r':  initialize = 1; pause = 0; break;
    case 'p':  pause = !pause; break;
    case 'k':  k0 = cur_num; initialize = 1; break;
    case 'l':  k1 = cur_num; initialize = 1; break;
    case '/':  draw_step = int(draw_step/2) + 1; break;
    case '*':  draw_step = int(draw_step*2); break;
    case '+':  ymag *= 1.2; break;
    case ';':  ymag /= 1.2; break;
    case '-':  ymag /= 1.2; break;
    default :  break;
    }
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y) { action((int) key); }

void k0_select(int value) { k0 = value; initialize = 1; }
void k1_select(int value) { k1 = value; initialize = 1; }

void make_menu(void)
{
    int K0, K1;
    K0 = glutCreateMenu(k0_select); {
        glutAddMenuEntry( "0", 0);
        glutAddMenuEntry( "1", 1);
        glutAddMenuEntry( "2", 2);
        glutAddMenuEntry( "3", 3);
        glutAddMenuEntry( "4", 4);
        glutAddMenuEntry( "5", 5);
        glutAddMenuEntry( "7", 7);
        glutAddMenuEntry("10",10);
        glutAddMenuEntry("15",15);
    }
    K1 = glutCreateMenu(k1_select); {
        glutAddMenuEntry( "0", 0);
        glutAddMenuEntry( "1", 1);
        glutAddMenuEntry( "2", 2);
        glutAddMenuEntry( "3", 3);
        glutAddMenuEntry( "4", 4);
        glutAddMenuEntry( "5", 5);
        glutAddMenuEntry( "7", 7);
        glutAddMenuEntry("10",10);
        glutAddMenuEntry("15",15);
    }
    glutCreateMenu(action); {
        glutAddMenuEntry("Pause/Continue",'p');
        glutAddMenuEntry("Restart",       'r');
	glutAddSubMenu("Large packet's K", K0);
        glutAddSubMenu("Small packet's K", K1);
        glutAddMenuEntry("Slow down",     '/');
        glutAddMenuEntry("Speed up",      '*');
        glutAddMenuEntry("Quit",          'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*****************************************
 * main()
 ----------------------------------------*/
int main(int argc, char** argv)
{
    /* init glut */
    glutInit (&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutCreateWindow(argv[0]);

    /* call-back */
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutKeyboardFunc(keyboard);
    make_menu();
    
    /* 2D-setup */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0., width - 1., 0., height - 1.);
    
    glutMainLoop ();
    return 0; 
}
